IMPLEMENTATION MODULE EGR;                                  
(**********************************************************************)
(*  PURPOSE   control linear EGR valve                                *)
(*                                                                    *)
(**********************************************************************)
(* Revisions By   Purpose                                             *)
(* 01/07/94  abc  added xTauFix                                       *)
(* 01/11/94  abc  xThrFix was xThrPos                                 *)
(* 01/12/94  abc  added ThrPos filter                                 *)
(* 01/28/94  abc  changed EGRMAP to MAP, moved it to In_out.imp       *)
(*                                                                    *)
(**********************************************************************)


FROM In_out IMPORT
  (* type *)  EngineModes,ThrDirs,
  (* var  *)  ClntTemp,RPM,MAP,EMAP,legrvi,IntPos,
  (* var  *)  ESADType,EngineMode,ThrDir,ThrPos,ThrReqI,
  (* var  *)  DACCAL,PedMax;

FROM Spark IMPORT
  (* var *)   SpkCtrlMax,SpkIClamp;

FROM DAC IMPORT
  (* var  *)  DAC;

VAR
  OldIntPos                                               : SHORTCARD;
  ModeTics                                                : CARDINAL;
  TargetOld,RPMOld,LegrOldErr,OTauFix,CTauFix             : LONGINT;
  Tau,RPMTics,FRPMDrv,LegrIntg,FltLift,LagTarget          : LONGINT;
  RPMComp,RPMDrv,BattVolts,PID,Duty,OpenLoop              : LONGINT;
  MAPComp,MinLegr,MaxLegr,OldLift,SpkComp,InitDuty        : LONGINT;
  MAPDrv,EMAPOld,CTempComp                                : LONGINT;
  OPLDuty,xOPLMult,FDuty,FMAP,IGClose                     : LONGINT;

CONST
  EGRLiftMax   : LONGINT  =   6144;  (* SF1 * 6MM lift                *)
  EGRLiftMin   : LONGINT  =    200;  (* SF1 * 0.2MM lift              *)
  EGRSlopeMax  : LONGINT  =     36;  (* max A/D counts per MM         *)
  EGRSlopeMin  : LONGINT  =     30;  (* min A/D counts per MM         *)
  OffZLmt      : CARDINAL =    400;  (* 5Msec ctr EngineMode(OffZ)    *)
  OffLmt       : CARDINAL =    400;  (* 5Msec ctr EngineMode(Off)     *)
  IdleZLmt     : CARDINAL =    400;  (* 5Msec ctr EngineMode(IdleZ)   *)
  IdleLmt      : CARDINAL =    400;  (* 5Msec ctr EngineMode(Idle)    *)
  ICloseLmt    : LONGINT  =    -50;  (* closing rate threshold, mm/s  *)
  SF1          : LONGINT  =   1024;  (* ScaleFactor #1                *)
  DIV0         : LONGINT  =      1;  (* to prevent DIV zero errors    *)
  One          : LONGINT  =   1024;  (* SF1 * 1                       *)
  Delta        : LONGINT  =      5;  (* SF1 * 5Msec per loop          *)
  BVNom        : LONGINT  =  13824;  (* SF1 * 13.5Volts               *)
  FLGain       : LONGINT  =     20;  (* SF1 * gain, 1024=no filter    *)
  BGain        : LONGINT  =    -41;  (* SF1 * -0.04                   *)
  MTStep       : LONGINT  =   1024;  (* SF1 * 10KPa/10                *)
  PGTStep      : LONGINT  =   1024;  (* SF1 * 1.0MM                   *)
  TargetStep   : LONGINT  =   1024;  (* SF1 * 1.0MM                   *)
  EFStep       : LONGINT  =    103;  (* SF1 * 0.1 Error fraction      *)
  MAPTabMult   : LONGINT  =    900;  (* SF1 * Initial Duty multiplier *)

  CTempMin     : LONGINT   =  45056;  (* SF1 * min coolant temp       *)
  CTempMax     : LONGINT   =  69632;  (* SF1 * max coolant temp       *)

  LiftMult     : LONGINT   =   1024;  (* SF1 * lift multiplier        *)
  LegrPGn      : LONGINT   =    100;  (* SF1 * PGain                  *)
  LegrIGn      : LONGINT   =   1000;  (* SF1 * IGain                  *)
  LegrDGn      : LONGINT   =    500;  (* SF1 * SF1 * DGain            *)
                                                                   
  SpkIGn       : LONGINT   =      0;  (* SF1 * IGain                  *)

  RPMTicLmt    : LONGINT   =     10;  (* Sec / 0.005                  *)
  RPMTau       : LONGINT   =    200;  (* SF1 * Sec                    *)
  RPMStep      : LONGINT   =    400;  (* 400 RPM                      *)
  RPMGain      : LONGINT   =    500;  (* SF1 * Gain                   *)

  ThrGainEnd   : LONGINT   =    409;  (* SF1 * 40% throttle           *)
  ThrRPMEnd    : LONGINT   =    204;  (* SF1 * 20% throttle           *)
  ThrRollOff   : LONGINT   =    716;  (* SF1 * 70% throttle           *)
  ThrStep2     : LONGINT   =     20;  (* SF1 *  2%                    *)
  ThrStep10    : LONGINT   =    102;  (* SF1 * 10%                    *)

  (* Proportional Control Gain *)
  PGainTab : ARRAY (0..6) OF LONGINT =                  (* SF1 * PGain *)
    ( 1024, 1024, 1024, 1024, 1024,  922,  819);
    (*----------------------------------------------------------------*)
    (* 0.0   1.0   2.0   3.0   4.0   5.0   6.0              Target MM *)
    (*----------------------------------------------------------------*)

  (* Integral Control Gain *)
  IGainTab : ARRAY (0..10),(0..6) OF LONGINT =          (* SF1 * IGain *)
    ((1024, 1024, 1024, 1024, 1024, 1024, 1024),     (* 0.0 ErrFrac   *)
     (1024, 1024, 1024, 1024, 1024, 1024, 1024),     (* 0.1           *)
     (1024, 1024, 1024, 1024, 1024, 1024, 1024),     (* 0.2           *)
     (1024, 1024, 1024, 1024, 1024, 1024, 1024),     (* 0.3           *)
     (1024, 1024,  980,  920,  720,  720,  720),     (* 0.4           *)
     (1024, 1024,  940,  840,  440,  440,  440),     (* 0.5           *)
     (1024, 1024,  920,  780,  200,  200,  200),     (* 0.6           *)
     (1024, 1024,  880,  700,  150,  150,  150),     (* 0.7           *)
     (1024, 1024,  860,  640,  100,  100,  100),     (* 0.8           *)
     (1024, 1024,  820,  560,   50,   50,   50),     (* 0.9           *)
     (1024, 1024,  800,  500,    0,    0,    0));    (* 1.0           *)
    (*----------------------------------------------------------------*)
    (* 0.0   1.0   2.0   3.0   4.0   5.0   6.0              Target MM *)
    (*----------------------------------------------------------------*)


PROCEDURE EGRCommand();
(* Commands EGR Valve Pintle Position *)

VAR
  I32a,I32b,I32c,Gain                                       : LONGINT;
  Alfa,RPMDelta,RPMAlfa,SSTarget                            : LONGINT;

BEGIN

  (************* open-loop, steady-state pintle lift ******************)
  IF (ThrPos < ThrGainEnd) THEN        (* look for table out of range *)
    I32a     := (Table_3D(ThrGainTab,ThrPos,RPM,ThrStep2,RPMStep));
    SSTarget := (xLiftMult * I32a) DIV SF1;
  ELSE
    SSTarget := EGRLiftMax;
  END;
  (* Roll-off pintle lift linearly from max lift at ThrRollOff to 
     zero at WOT. *)
  IF (ThrPos > ThrRollOff) THEN
    SSTarget := ((EGRLiftMax*(One-ThrPos))DIV(DIV0 + One - ThrRollOff));
  END;

  (****************** lag-lead filter *********************************)
  (* Lag-lead filter the steady-state lift command to match manifold
     pressure and mixing dynamics. *)
  IF (ThrDir = Open) THEN  (* Assign time constant, Tau, for opening. *)
    Tau := OTauFix;                               (* Tau=1024=1.0 sec *)            
    (* Tau := Table_3D(ThrOpenTab,ThrPos,RPM,ThrStep10,RPMStep); *)
  ELSIF(ThrDir=Close)THEN        (* Assign time constant for closing. *)
    Tau := CTauFix;
    (* Tau := Table_3D(ThrCloseTab,ThrPos,RPM,ThrStep10,RPMStep); *)
  END;
  (* Calculate lag-lead filter parameters. *)
  Alfa  := ((SF1 * Delta) DIV (DIV0 + Tau));
  IF (Alfa > One) THEN
    Alfa := One;
  END;
  I32c      := (One - Alfa);
  I32b      := xBGain;
  I32a      := (One - I32b - I32c);
  (* Perform filtering. *)
  LagTarget := (((I32a*SSTarget)+(I32b*TargetOld)+(I32c*LagTarget))
               DIV SF1);
  TargetOld := SSTarget;                              (* 1024 = 1.0MM *)

  (*********** Adjust target lift for rate of change of RPM ***********)
  (* Calculate rate of change of rpm. *)      
  INC(RPMTics);
  IF(RPMTics = RPMTicLmt) THEN
    RPMDelta := Delta * RPMTicLmt;                   (* 1024 = 1.0Sec *)
    RPMDrv   := (SF1*(RPM-RPMOld)) DIV (DIV0 + RPMDelta);
    RPMOld   := RPM;                              (* 1024=1024RPM/Sec *)
    (* Lag filter the rpm rate. *)
    RPMAlfa  := (SF1 * RPMDelta) DIV (DIV0 + xRPMTau);  (* 1024 = 1.0 *)
    IF (RPMAlfa > One) THEN
      RPMAlfa := One;
    END;
    FRPMDrv  := (RPMAlfa * RPMDrv+ (One-RPMAlfa)*FRPMDrv) DIV SF1;
    (* Assign rpm compensator gain. *)
    IF (ThrPos < ThrRPMEnd) THEN      (* Look for table out of range. *)
      Gain := (xRPMGain*Table_2D(RPMGainTab,ThrPos,ThrStep2)) DIV SF1;
    ELSE
      Gain   := (xRPMGain * 20) DIV SF1;
    END;
    IF (xRPMGainFix >= 0) THEN           (* Fix the compensator gain. *)
      Gain := xRPMGainFix;
    END;
    (* Calculate lift correction. *)
    RPMComp := One - (Gain * FRPMDrv) DIV 1000;         (* 1024 = 1.0 *)
    RPMTics := 0;
  END;

  (********** Open-Loop EGR Lift Command ******************************)
  OpenLoop := (LagTarget * RPMComp) DIV SF1;

  (**** Closed-Loop EGR Control Using MBT Spark Timing Target *********)
  IF (SpkIClamp=FALSE) OR ((SpkIClamp=TRUE)AND(SpkCtrlMax<0)) THEN
    I32a := (xSpkIGn * SpkCtrlMax) DIV SF1;                 (* SF1**1 *)
    IF ((I32a > 0) AND (EGRTarget < EGRLiftMax)) OR
       ((I32a < 0) AND (EGRTarget > 0)) THEN
      SpkComp := (SpkComp + I32a);                          (* SF1**1 *)
    END;
  END;
  IF (xSpkIGn = 0) THEN
    SpkComp := 0;
  END;

  (********************** coolant temp compensation *******************)
  CTempComp := ((ClntTemp-CTempMin) * SF1) DIV (DIV0+CTempMax-CTempMin);
  IF (CTempComp < 0) THEN
    CTempComp := 0;
  ELSIF (CTempComp > One) THEN
    CTempComp := One;
  END;

  (*********** Determine EGR Valve Lift Target ************************)
  IF (xEGRFix >= 0) THEN                       (* Fix the target lift *)
    EGRTarget := xEGRFix;
  ELSE
    EGRTarget := ((CTempComp * OpenLoop) DIV SF1) + SpkComp;
  END;
  (* Limit the lift command. *)
  IF (EGRTarget < 0) THEN
    EGRTarget := 0;
  ELSIF (EGRTarget > 6144) THEN
    EGRTarget := 6144;
  END;
  (* At WOT, command zero egr and reset the spark control integral *)
  IF (LONGINT (ThrReqI) > PedMax) THEN          (* Zero EGR at WOT *)
     EGRTarget := 0;
     SpkComp := 0 ;
  END;

END EGRCommand;



(* Initialize all variables that are not calculated prior
   first use. *)
PROCEDURE EGRInit();
  (* initialize EGR code *)
BEGIN
  EGREnable  :=  TRUE;
  EGRTarget  :=     0;
  EGRDuty    :=     0;
  EGRLift    :=     0;
  EGRLZero   :=  1000;                (* nominal value for EGRLZero   *)
  EGRMinDuty :=   400;                (* nominal value for EGRMinDuty *)
  FDuty      :=   450;
  EGRMinMAP  :=  3000;                 (* nominal value for EGRMinMap *)
  FMAP       :=  3000;
  EGRSlope   :=    33;                  (* nominal value for EGRSlope *)
  FltLift    :=   230;                   (* nominal value for MaxLegr *)
  MAPComp    :=    40;
  MAPDrv     :=     0;
  EMAPOld    :=     0;
  EngineMode :=   Off;
  OldLift    :=     0;
  RPMOld     :=     0;
  RPMComp    :=  1024;
  RPMDrv     :=     0;
  SpkComp    :=     0;
  CTempComp  :=  1024;
  BattVolts  := 13824;                             (* SF1 * 13.5Volts *)
  ModeTics   :=     0;       (* initialize timer for EngineMode = Off *)
  LegrIntg   :=     0;
  LegrOldErr :=     0;
  OTauFix    :=   100;
  CTauFix    :=   100;
  RPMTics    :=     0;
  FRPMDrv    :=     0;
  TargetOld  :=     0;
  OldIntPos  :=     0;
  IGClose    :=  1024;                    (* Closing IGain Multiplier *)


  xDutyFix     :=       -1;
  xEGRFix      :=        0;
  xFLGain      :=   FLGain;    
  xBGain       :=    BGain;    
  xLiftMult    := LiftMult;    
  xLegrPGn     :=  LegrPGn;    
  xLegrIGn     :=  LegrIGn;    
  xLegrDGn     :=  LegrDGn;    
  xMAPTabMult  := MAPTabMult;  
  xSpkIGn      :=   SpkIGn;    
  xRPMTau      :=   RPMTau;    
  xRPMGain     :=  RPMGain;    
  xRPMGainFix  :=       -1;    
  xPGnFix      :=        0;    
  xIGnFix      :=        0;    
  xAdjMult     :=     1130;    
  xLead        :=      250;    
  xOPLMult     :=      800;    
END EGRInit;


END EGR.



